'From Squeak3.10beta of 22 July 2007 [latest update: #7158] on 23 December 2007 at 10:54:14 pm'!
Object subclass: #AsQLTuples
	instanceVariableNames: 'arity contents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Small-QLog'!
!AsQLTuples commentStamp: 'kwl 12/22/2007 18:14' prior: 0!
Instances of AsQLTuples aim to provide Datalog query program semantics which allows relational algebra with cross-product, union, selection, intersection, projection and set-difference.

	toDo: provide #distinct (tuples), also #rehashColumn: and #exists:column: access
	toDo: grouping and aggregation, looks to be easy in Smalltalk :)
	toDo: recursion (iteration until least fixpoint is found, preferably with worklist algorithm)

In this incarnation a query program can be composed from four components:

o	a relation (or cross-product from an array of collections, any of which can be a block)
o	where block (optional) for intersecting the relation resp. selecting tuples from it
o	select block (optional) for projecting and/or replacing columns resp. tuples
o	do block (optional) for enumerating the tuples of the solution

The solution of a query program can be used as initial relation in another query program.

Example: instance and class methods of AsQLTuples
(
 | solution | solution :=
 (AsQLTuples from: {{AsQLTuples}, {AsQLTuples class}. [:c | c selectors asSortedCollection]})
	"declares a relation with arity 2, a cross-product of classes and what they implement"

   select: [:c :s | {MethodReference new setStandardClass: c methodSymbol: s}];
	"the relation is now projected and the remaining column is replaced"

   contents.
 SystemNavigation default browseMessageList: solution name: 'AsQLTuples methods'
)

Example: what do AsQLTuples methods do with Streams+Arrays
(
 | solution support |
 support := (support := (Array withAllSuperclasses, WriteStream withAllSuperclasses) copyWithout: Object)
	, (support collect: [:each | each class]).

 solution := ((AsQLTuples from: {{AsQLTuples}, {AsQLTuples class}.
		[:c | c selectors].
		[:c :s | (c compiledMethodAt: s) messages]})
	"declares a relation with arity 3, a cross-product of classes, what they implement+send"

   where: [:c :s :m | (Integer canUnderstand: m) not];
	"the relation is now selected upon another relation (here a predicate)"

   select: [:c :s :m | {m}];
	"the relation is now projected (one column remains)"

   contents) asIdentitySet. solution :=
 (AsQLTuples from: {support. [:c | c selectors asSortedCollection]})
	"declares a relation with arity 2, a cross-product of classes and what they implement"

   where: [:c :s | solution includes: s];
	"the relation is now selected upon another relation (here a predicate)"

   select: [:c :s | {MethodReference new setStandardClass: c methodSymbol: s}];
	"the relation is now projected and the remaining column is replaced"

   contents.
 SystemNavigation default browseMessageList: solution name: 'what do AsQLTuples methods do with Streams+Arrays'
)

Example: Smalltalk methods which ought to be fast accessors
(
 | solution | solution :=
 (AsQLTuples from: {Smalltalk allClasses. [:c | c selectors]})
	"declares a relation with arity 2, a cross-product of classes and what they implement"

   where: [:c :s | | m | (m := c compiledMethodAt: s) isQuick and: [m primitive > 264]];
	"the relation is now selected upon another relation (here a predicate)"

   select: [:c :s | {MethodReference new setStandardClass: c methodSymbol: s}];
	"the relation is now projected and the remaining column is replaced"

   contents.
 SystemNavigation default browseMessageList: solution name: 'quick field accessors'
)

The implementation of AsQLTuples was inspired in part by
-	http://bddbddb.sourceforge.net/
-	http://progtools.comlab.ox.ac.uk/projects/codequest
-	http://semmle.com/documentation/further-reading/
!


!AsQLTuples methodsFor: 'initialize-release' stamp: 'kwl 12/22/2007 15:27'!
initializeWith: anArray
	"initialize the receiver from the argument"

	contents := Array new: (arity := anArray size).
	1 to: arity do: [:ix | contents at: ix put: (Array new: ix)].
	"compute cross-product from the collections in the argument"
	contents := Array streamContents: [:newContents |
		self initializeColumn: 1 from: anArray to: newContents]! !


!AsQLTuples methodsFor: 'accessing' stamp: 'kwl 12/22/2007 16:42'!
arity
	^ arity! !

!AsQLTuples methodsFor: 'accessing' stamp: 'kwl 12/22/2007 16:42'!
contents
	^ contents! !


!AsQLTuples methodsFor: 'operations' stamp: 'kwl 12/22/2007 16:13'!
select: aBlock 
	"let aBlock determine the receiver's contents"
	| tuple zw numTuples |
	tuple := Array new: arity.
	numTuples := 0.
	contents := Array streamContents: [:newContents |
		1 to: contents size by: arity do: [:ix | 
			zw := ix - 1.
			1 to: arity do: [:jy |
				tuple at: jy put: (contents at: (zw := 1 + zw))].
			newContents nextPutAll: (aBlock valueWithArguments: tuple).
			numTuples := 1 + numTuples]].
	arity := contents size // numTuples! !

!AsQLTuples methodsFor: 'operations' stamp: 'kwl 12/22/2007 16:04'!
where: aBlock 
	"let aBlock select the receiver's tuples"
	| tuple wz |
	tuple := Array new: arity.
	contents := Array streamContents: [:newContents |
		1 to: contents size by: arity do: [:ix | 
			wz := ix - 1.
			1 to: arity do: [:jy |
				tuple at: jy put: (contents at: (wz := 1 + wz))].
			(aBlock valueWithArguments: tuple)
				ifTrue: [newContents nextPutAll: tuple]]]! !


!AsQLTuples methodsFor: 'enumeration' stamp: 'kwl 12/22/2007 16:09'!
do: aBlock 
	"Evaluate aBlock with each of the receiver's tuples as argument."
	| tuple zw |
	tuple := Array new: arity.
	1 to: contents size by: arity do: [:ix | 
		zw := ix - 1.
		1 to: arity do: [:jy |
			tuple at: jy put: (contents at: (zw := 1 + zw))].
		aBlock valueWithArguments: tuple]! !


!AsQLTuples methodsFor: 'private' stamp: 'kwl 12/22/2007 16:49'!
initializeColumn: ix from: argV to: newContents 
	| tuple jy column clone remainder |
	tuple := contents at: ix.
	(jy := 1 + ix) <= arity
		ifTrue: [(column := argV at: jy) isBlock
			ifTrue: [clone := argV clone]].
	"recursively compute cross-product"
	(argV at: ix) value do: [:element | 
		ix = arity
			ifTrue: ["at the last column"
				newContents nextPutAll: (tuple at: ix put: element; yourself)]
			ifFalse: ["at an intermediate column"
				column isBlock
					ifTrue: ["ask block for elements"
						(remainder := clone) at: jy
							put: (column valueWithArguments:
								(tuple at: ix put: element; yourself))]
					ifFalse: [remainder := argV].
				"append element to remaining rows"
				jy to: arity do: [:wz | (contents at: wz) at: ix put: element].
				self initializeColumn: jy from: remainder to: newContents]]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

AsQLTuples class
	instanceVariableNames: ''!

!AsQLTuples class methodsFor: 'instance creation' stamp: 'kwl 12/22/2007 14:53'!
from: anArray
	"answer an initialized instance of the receiver"

	^ self basicNew initializeWith: anArray! !
